import os
import re
import tempfile
import logging
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import (IdentityInterface, Function, Merge)
from nipype.interfaces.ants import (AverageImages, Registration,
                                    ApplyTransforms)
from nipype.interfaces import fsl
from nipype.interfaces.dcmstack import CopyMeta
import qiutil
from qiutil.logging import logger
from qixnat import project
from ..interfaces import (Copy, XNATUpload)
from ..helpers import bolus_arrival
from .workflow_base import WorkflowBase


REG_PREFIX = 'reg'
"""The XNAT registration resource name prefix."""


def run(subject, session, bolus_arrival_index, *images, **opts):
    """
    Runs the registration workflow on the given session scan images.

    :param subject: the subject name
    :param session: the session name
    :param images: the input session scan images
    :param mask: the image mask
    :param bolus_arrival_index: the required bolus uptake series index
    :param opts: the :class:`RegistrationWorkflow` initializer
        and :meth:`RegistrationWorkflow.run` options
    :return: the realigned image file path array
    """
    # Extract the run options.
    run_opts = {k: opts.pop(k) for k in ['dest', 'mask'] if k in opts}
    # Make the workflow.
    reg_wf = RegistrationWorkflow(**opts)
    # Run the workflow.
    return reg_wf.run(subject, session, bolus_arrival_index, *images,
                      **run_opts)


def generate_resource_name():
    """
    Makes a unique registration resource name. Uniqueness permits more than
    one registration to be stored for a given session without a name conflict.

    :return: a unique XNAT registration resource name
    """
    return "%s_%s" % (REG_PREFIX, qiutil.file.generate_file_name())


class RegistrationWorkflow(WorkflowBase):

    """
    The RegistrationWorkflow class builds and executes the registration workflow.
    The workflow registers an input NiFTI scan image against the input reference
    image and uploads the realigned image to XNAT.

    The registration workflow input is the *input_spec* node consisting of the
    following input fields:

    - *subject*: the subject name

    - *session*: the session name

    - *mask*: the mask to apply to the images

    - *reference*: the fixed reference image

    - *image*: the image file to register

    The mask can be obtained by running the
    :class:`qipipe.pipeline.mask.MaskWorkflow` workflow.

    The reference can be obtained by running the
    :class:`qipipe.pipeline.reference.ReferenceWorkflow` workflow.

    The output realigned image file is named
    *scan*\ ``_``\ *resource*\ ``.nii.gz``, where *scan* is the input
    image file name without extension and *resource* is the XNAT
    registration resource name

    Three registration techniques are supported:

    - ``ants``: ANTS_ SyN_ symmetric normalization diffeomorphic registration
      (default)

    - ``fnirt``: FSL_ FNIRT_ non-linear registration
    
    - ``mock``: Test technique which copies each input scan image to
      the output image file

    The optional workflow configuration file can contain overrides for the
    Nipype interface inputs in the following sections:

    - ``ants.Registration``: the ANTS `Registration interface`_ options

    - ``ants.ApplyTransforms``: the ANTS `ApplyTransform interface`_ options

    - ``fsl.FNIRT``: the FSL `FNIRT interface`_ options

    :Note: Since the XNAT *resource* name is unique, a
        :class:`qipipe.pipeline.registration.RegistrationWorkflow` instance
        can be used for only one registration workflow. Different registration
        inputs require different
        :class:`qipipe.pipeline.registration.RegistrationWorkflow` instances.

    .. _ANTS: http://stnava.github.io/ANTs/
    .. _ApplyTransform interface: http://nipy.sourceforge.net/nipype/interfaces/generated/nipype.interfaces.ants.resampling.html
    .. _FNIRT: http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FNIRT#Research_Overview
    .. _FNIRT interface: http://nipy.sourceforge.net/nipype/interfaces/generated/nipype.interfaces.fsl.preprocess.html
    .. _FSL: http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FSL
    .. _Registration interface: http://nipy.sourceforge.net/nipype/interfaces/generated/nipype.interfaces.ants.registration.html
    .. _SyN: http://www.ncbi.nlm.nih.gov/pubmed/17659998
    """

    def __init__(self, **opts):
        """
        If the optional configuration file is specified, then the workflow
        settings in that file override the default settings.

        :param opts: the following initialization options:
        :keyword base_dir: the workflow execution directory
            (default a new temp directory)
        :keyword cfg_file: the optional workflow inputs configuration file
        :keyword resource: the XNAT resource name to use
        :keyword technique: the case-insensitive workflow technique
            (``ANTS`` or ``FNIRT``, default ``ANTS``)
        """
        cfg_file = opts.pop('cfg_file', None)
        super(RegistrationWorkflow, self).__init__(logger(__name__), cfg_file)

        rsc = opts.pop('resource', None)
        if not rsc:
            rsc = generate_resource_name()
        self.resource = rsc
        """The XNAT resource name used for all runs against this workflow
        instance."""

        self.workflow = self._create_realignment_workflow(**opts)
        """The registration realignment workflow."""

    def run(self, subject, session, bolus_arrival_index, *images, **opts):
        """
        Runs the registration workflow on the given session scan images.

        :param subject: the subject name
        :param session: the session name
        :param bolus_arrival_index: the required bolus uptake series index
        :param images: the input session scan images
        :param opts: the following options:
        :option mask: the image mask file path
        :option dest: the realigned image target directory (default is the
            current directory)
        :return: the realigned output file paths
        """
        if not images:
            return []
        # Sort the images by series number.
        sorted_scans = sorted(images)

        # Split the images into pre- and post-arrival.
        pre_arrival = sorted_scans[0: bolus_arrival_index + 1]
        reg_input = sorted_scans[bolus_arrival_index + 1:]
        
        # The initial reference image occurs at bolus arrival.
        ref_0_image = pre_arrival[bolus_arrival_index]

        # The target location.
        dest = opts.get('dest', None)
        if dest:
            dest = os.path.abspath(dest)
        else:
            dest = os.getcwd()

        # The execution workflow.
        exec_wf = self._create_execution_workflow(bolus_arrival_index,
                                                  ref_0_image, dest)

        # Set the execution workflow inputs.
        input_spec = exec_wf.get_node('input_spec')
        input_spec.inputs.subject = subject
        input_spec.inputs.session = session
        input_spec.inputs.pre_arrival = pre_arrival
        mask = opts.get('mask', None)
        if mask:
            input_spec.inputs.mask = mask

        # Iterate over the registration inputs.
        iter_reg_input = exec_wf.get_node('iter_reg_input')
        iter_reg_input.iterables = ('image', reg_input)

        # Execute the workflow.
        self._logger.debug("Executing the %s workflow on %s %s..." %
                         (self.workflow.name, subject, session))
        self._run_workflow(exec_wf)
        self._logger.debug("Executed the %s workflow on %s %s." %
                         (self.workflow.name, subject, session))

        # Return the output files.
        return [os.path.join(dest, filename(scan)) for scan in images]

    def _create_execution_workflow(self, bolus_arrival_index, ref_0_image,
                                   dest):
        """
        Makes the registration execution workflow on the given session
        scan images.

        The execution workflow input is the *input_spec* node consisting of the
        following input fields:

        - *subject*: the subject name

        - *session*: the session name

        - *mask*: the mask to apply to the images

        - *pre_arrival*: the scan images prior to bolus arrival, which will
          not be realigned.

        - *reference*: the fixed reference for the given image registration
        
        In addition, the caller has the responsibility of setting the
        ``iter_reg_input`` iterables to the scans to realign.

        The *reference* input is set by :meth:`connect_reference`.

        :param bolus_arrival_index: the required bolus uptake series index
        :param ref_0_image: the required initial fixed reference image
        :param dest: the required target realigned image directory
        :return: the execution workflow
        """
        if bolus_arrival_index == None:
            raise ValueError('Registration workflow is missing the bolus' +
                             ' arrival index')
        if not ref_0_image:
            raise ValueError('Registration workflow is missing the initial' +
                             ' fixed reference image')
        if not dest:
            raise ValueError('Registration workflow is missing the destination' +
                             ' directory')
        self._logger.debug("Creating the registration execution workflow"
                           " with bolus arrival index %d and initial reference"
                           " %s..." % (bolus_arrival_index, ref_0_image))

        # The execution workflow.
        exec_wf = pe.Workflow(name='reg_exec', base_dir=self.workflow.base_dir)

        # The registration workflow input.
        input_fields = ['subject', 'session', 'pre_arrival', 'mask', 'ref_0', 'resource']
        input_spec = pe.Node(IdentityInterface(fields=input_fields),
                             name='input_spec')
        input_spec.inputs.ref_0 = ref_0_image
        exec_wf.connect(input_spec, 'subject',
                        self.workflow, 'input_spec.subject')
        exec_wf.connect(input_spec, 'session',
                        self.workflow, 'input_spec.session')
        exec_wf.connect(input_spec, 'mask',
                        self.workflow, 'input_spec.mask')
        input_spec.inputs.resource = self.resource

        # The input images are iterable. The reference is set by the
        # connect_reference method below.
        iter_reg_input = pe.Node(IdentityInterface(fields=['image', 'reference']),
                             name='iter_reg_input')
        exec_wf.connect(iter_reg_input, 'image',
                        self.workflow, 'input_spec.moving_image')
        exec_wf.connect(iter_reg_input, 'reference',
                        self.workflow, 'input_spec.reference')

        # The output destination directory.
        if not os.path.exists(dest):
            os.makedirs(dest)

        # Copy the pre-arrival files to the destination directory.
        copy_pre_arrival_func = Function(
            input_names=['in_files', 'dest'],
            output_names=['out_files'], function=copy_files)
        copy_pre_arrival = pe.Node(copy_pre_arrival_func, dest=dest,
                                   name='copy_pre_arrival')
        # Work around the following Nipype bug:
        # * the Function Node input is not set by the Node constructor
        if not copy_pre_arrival.inputs.dest:
            copy_pre_arrival.inputs.dest = dest
        exec_wf.connect(input_spec, 'pre_arrival', copy_pre_arrival, 'in_files')

        # Copy the realigned image to the destination directory.
        copy_realigned = pe.Node(Copy(dest=dest), name='copy_realigned')
        exec_wf.connect(self.workflow, 'output_spec.out_file',
                        copy_realigned, 'in_file')

        # Set the recursive realigned -> reference connections.
        connect_ref_args = dict(bolus_arrival_index=bolus_arrival_index,
                                initial_reference=ref_0_image)
        exec_wf.connect_iterables(copy_realigned, iter_reg_input,
                                  connect_reference, **connect_ref_args)

        # Collect the realigned images.
        join_realigned = pe.JoinNode(IdentityInterface(fields=['images']),
                                     joinsource='iter_reg_input',
                                     joinfield='images',
                                     name='join_realigned')
        exec_wf.connect(copy_realigned, 'out_file', join_realigned, 'images')

        # Merge the pre-arrival scans and the realigned images.
        concat_reg = pe.Node(Merge(2), name='concat_reg')
        exec_wf.connect(copy_pre_arrival, 'out_files', concat_reg, 'in1')
        exec_wf.connect(join_realigned, 'images', concat_reg, 'in2')

        # Upload the registration result into the XNAT registration resource.
        upload_reg_xfc = XNATUpload(project=project(),
                                    resource=self.resource)
        upload_reg = pe.Node(upload_reg_xfc, name='upload_reg')
        exec_wf.connect(input_spec, 'subject', upload_reg, 'subject')
        exec_wf.connect(input_spec, 'session', upload_reg, 'session')
        exec_wf.connect(concat_reg, 'out', upload_reg, 'in_files')

        # The execution output.
        output_spec = pe.Node(IdentityInterface(fields=['images']),
                              name='output_spec')
        exec_wf.connect(concat_reg, 'out', output_spec, 'images')

        self._logger.debug("Created the %s workflow." % exec_wf.name)
        # If debug is set, then diagram the workflow graph.
        if self._logger.level <= logging.DEBUG:
            self.depict_workflow(exec_wf)

        return exec_wf

    def _create_realignment_workflow(self, **opts):
        """
        Creates the workflow which registers and resamples images.
        The registration workflow performs the following steps:

        - Generates a unique XNAT resource name

        - Set the mask and realign workflow inputs

        - Run these workflows

        - Upload the realign outputs to XNAT
        
        :param opts: the following workflow options:
        :keyword base_dir: the workflow execution directory
            (default is a new temp directory)
        :keyword technique: the registration technique
            (``ants``, `fnirt`` or ``mock``, default ``ants``)
        :return: the Workflow object
        """
        self._logger.debug("Creating the registration realignment workflow...")

        # The workflow.
        base_dir = opts.get('base_dir', tempfile.mkdtemp())
        realign_wf = pe.Workflow(name='registration', base_dir=base_dir)

        # The workflow input.
        in_fields = ['subject', 'session', 'moving_image', 'reference',
                     'mask', 'resource']
        input_spec = pe.Node(IdentityInterface(fields=in_fields),
                             name='input_spec')

        # The realign workflow.
        technique = opts.get('technique')
        if not technique:
            technique = 'ants'
        input_spec.inputs.resource = self.resource

        # Copy the DICOM meta-data. The copy target is set by the technique
        # node defined below.
        copy_meta = pe.Node(CopyMeta(), name='copy_meta')
        realign_wf.connect(input_spec, 'moving_image', copy_meta, 'src_file')

        if technique.lower() == 'ants':
            # Nipype bug work-around:
            # Setting the registration metric and metric_weight inputs after the
            # node is created results in a Nipype input trait dependency warning.
            # Avoid this warning by setting these inputs in the constructor
            # from the values in the configuration.
            reg_cfg = self._interface_configuration(Registration)
            metric_inputs = {field: reg_cfg[field]
                             for field in ['metric', 'metric_weight']
                             if field in reg_cfg}
            # Register the images to create the rigid, affine and SyN
            # ANTS transformations.
            register = pe.Node(Registration(**metric_inputs), name='register')
            realign_wf.connect(input_spec, 'reference',
                               register, 'fixed_image')
            realign_wf.connect(input_spec, 'moving_image',
                               register, 'moving_image')
            realign_wf.connect(input_spec, 'mask',
                               register, 'fixed_image_mask')
            realign_wf.connect(input_spec, 'mask',
                               register, 'moving_image_mask')
            # Get the file name without directory.
            input_filename_xfc = Function(input_names=['in_file'],
                                          output_names=['out_file'],
                                          function=filename)
            input_filename = pe.Node(input_filename_xfc, name='input_filename')
            realign_wf.connect(input_spec, 'moving_image',
                               input_filename, 'in_file')
            # Apply the transforms to the input image.
            apply_xfm = pe.Node(ApplyTransforms(), name='apply_xfm')
            realign_wf.connect(input_spec, 'reference',
                               apply_xfm, 'reference_image')
            realign_wf.connect(input_spec, 'moving_image',
                               apply_xfm, 'input_image')
            realign_wf.connect(input_filename, 'out_file',
                               apply_xfm, 'output_image')
            realign_wf.connect(register, 'forward_transforms',
                               apply_xfm, 'transforms')
            # Copy the meta-data.
            realign_wf.connect(apply_xfm, 'output_image',
                               copy_meta, 'dest_file')
        elif technique.lower() == 'fnirt':
            # Copy the input to a work directory, since FNIRT adds
            # temporary files to the input image location.
            fnirt_copy_moving = pe.Node(Copy(), name='fnirt_copy_moving')
            realign_wf.connect(input_spec, 'moving_image',
                               fnirt_copy_moving, 'in_file')
            # Get the affine transformation.
            flirt = pe.Node(fsl.FLIRT(), name='flirt')
            realign_wf.connect(input_spec, 'reference', flirt, 'reference')
            realign_wf.connect(fnirt_copy_moving, 'out_file', flirt, 'in_file')
            # Register the image.
            fnirt = pe.Node(fsl.FNIRT(), name='fnirt')
            realign_wf.connect(input_spec, 'reference', fnirt, 'ref_file')
            realign_wf.connect(flirt, 'out_matrix_file', fnirt, 'affine_file')
            realign_wf.connect(fnirt_copy_moving, 'out_file', fnirt, 'in_file')
            realign_wf.connect(input_spec, 'mask', fnirt, 'inmask_file')
            realign_wf.connect(input_spec, 'mask', fnirt, 'refmask_file')
            # Copy the meta-data.
            realign_wf.connect(fnirt, 'warped_file', copy_meta, 'dest_file')
        elif technique.lower() == 'mock':
            # Copy the input scan file to an output file.
            mock_copy = pe.Node(Copy(), name='mock_copy')
            realign_wf.connect(input_spec, 'moving_image',
                               mock_copy, 'in_file')
            realign_wf.connect(mock_copy, 'out_file', copy_meta, 'dest_file')
        else:
            raise ValueError("Registration technique not recognized: %s" %
                             technique)

        # The output is the realigned image.
        output_spec = pe.Node(IdentityInterface(fields=['out_file']),
                              name='output_spec')
        realign_wf.connect(copy_meta, 'dest_file', output_spec, 'out_file')

        self._configure_nodes(realign_wf)

        self._logger.debug("Created the %s workflow." % realign_wf.name)
        # If debug is set, then diagram the workflow graph.
        if self._logger.level <= logging.DEBUG:
            self.depict_workflow(realign_wf)
        return realign_wf


### Utility functions called by the workflow nodes. ###

def filename(in_file):
    """
    :param in_file: the input file path
    :return: the file name without a directory
    """
    import os

    return os.path.split(in_file)[1]


def copy_files(in_files, dest):
    """
    :param in_files: the input files
    :param dest: the destination directory
    :return: the output files
    """
    from qipipe.interfaces import Copy
    
    return [Copy(in_file=in_file, dest=dest).run().outputs.out_file
            for in_file in in_files]


def connect_reference(workflow, realigned_nodes, input_nodes,
                      bolus_arrival_index, initial_reference):
    """
    Connects the reference input for the given reference workflow input nodes
    as follows:

    * Prior to bolus arrival, the reference is the successor realignment.

    * The reference for the nodes before and after the given initial reference
      node is that initial reference.

    * The bolus arrival successor node reference is the *initial_reference*.

    The initial reference is the input node for the scan immediately
    following bolus uptake. This node is not included in the iterable
    expansion nodes.

    :param workflow: the workflow delegate which connects nodes
    :param ref_nodes: the iterable expansion reference input nodes
    :param realigned_nodes: the iterable expansion realignment output nodes
    :param bolus_arrival_index: the bolus uptake series index
    :param initial_reference: the starting reference input node
    """
    # The number of input and realigned nodes.
    node_cnt = len(input_nodes)
    # The reference is the successor realignment up to the bolus arrival.
    end = min(bolus_arrival_index - 1, node_cnt - 1)
    for i in range(0, end):
        workflow.connect(realigned_nodes[i + 1], 'out_file',
                         input_nodes[i], 'reference')

    # The reference for the bolus arrival predecessor and successor
    # is the initial reference.
    start = max(bolus_arrival_index - 1, 0)
    end = min(bolus_arrival_index + 1, node_cnt)
    for i in range(start, end):
        input_nodes[i].inputs.reference = initial_reference

    # The reference is the predecessor realignment for the remaining
    # nodes.
    for i in range(bolus_arrival_index, node_cnt - 1):
        workflow.connect(realigned_nodes[i], 'out_file',
                         input_nodes[i + 1], 'reference')



import os
import re
import glob
import shutil
from nose.tools import (assert_true, assert_is_not_none)
import nipype.pipeline.engine as pe
from nipype.interfaces.dcmstack import MergeNifti
from qipipe.pipeline import (mask, qipipeline)
from ... import (project, ROOT)
from ...helpers.logging import logger
from ...unit.pipeline.staged_test_base import StagedTestBase

FIXTURES = os.path.join(ROOT, 'fixtures', 'registration')
"""The test fixtures directory."""

MASK_CONF = os.path.join(ROOT, 'conf', 'mask.cfg')
"""The test mask configuration."""

RESULTS = os.path.join(ROOT, 'results', 'pipeline', 'mask')
"""The test results directory."""


class TestMaskWorkflow(StagedTestBase):
    """
    Mask workflow unit tests.
    
    This test exercises the mask workflow on three series of one visit in each
    of the Breast and Sarcoma studies.
    
    :Note: This test is disabled pending a time series test fixture.
     
     TODO - Make a time series test fixture and enable the tests.
    """

    def __init__(self):
        super(TestMaskWorkflow, self).__init__(logger(__name__), FIXTURES,
                                               RESULTS)


    def tearDown(self):
        super(TestMaskWorkflow, self).tearDown()

    def test_breast(self):
        self._test_breast()
    
    def test_sarcoma(self):
        self._test_sarcoma()

    def _run_workflow(self, subject, session, *images, **opts):
        """
        Executes :meth:`qipipe.pipeline.mask.run` on the given input.
        
        :param subject: the input subject
        :param session: the input session
        :param time_series: the input time series
        :param opts: the :class:`qipipe.pipeline.modeling.MaskWorkflow`
            initializer options
        :return: the :meth:`qipipe.pipeline.mask.run` result
        """
        # Make the 4D time series from the test fixture inputs.
        merge = MergeNifti(in_files=list(images),
                           out_format=qipipeline.SCAN_TS_RSC)
        time_series = merge.run().outputs.out_file
        logger(__name__).debug("Testing the mask workflow on the %s %s time"
                               " series %s..." %
                               (subject, session, time_series))
        
        return mask.run(subject, session, time_series, cfg_file=MASK_CONF,
                        **opts)

    def _verify_result(self, xnat, subject, session, result):
        # Verify that the mask XNAT resource was created.
        rsc_obj = xnat.find(project(), subject, session, resource=result)
        assert_is_not_none(rsc_obj, "The %s %s XNAT mask resource object was"
                                    " not created" % (subject, session))


if __name__ == "__main__":
    import nose

    nose.main(defaultTest=__name__)



import os
import glob
from nose.tools import assert_equal
from qipipe.staging.staging_helper import *
from qiutil.logging import logger
from ... import ROOT

# The test fixture.
FIXTURE = os.path.join(ROOT, 'fixtures', 'staging', 'sarcoma', 'Subj_1')


class TestStagingHelper(object):

    """staging_helper unit tests."""

    def test_group_dicom_files_by_series(self):
        dicom_files = glob.glob(FIXTURE + '/V*/*concat*/*')
        groups = group_dicom_files_by_series(*dicom_files)
        assert_equal(set(groups.keys()), set(
            [9, 10]), "The DICOM series grouping is incorrect: %s" % groups)


if __name__ == "__main__":
    import nose
    nose.main(defaultTest=__name__)



from nipype.interfaces.base import (traits,
                                    BaseInterfaceInputSpec, TraitedSpec, BaseInterface,
                                    InputMultiPath, OutputMultiPath, Directory, File)
from qidicom.hierarchy import group_dicom_files_by_series


class GroupDicomInputSpec(BaseInterfaceInputSpec):
    in_files = InputMultiPath(
        traits.Either(File(exists=True), Directory(exists=True)),
        mandatory=True, desc='The DICOM files to group')


class GroupDicomOutputSpec(TraitedSpec):
    series_files_dict = traits.Dict(desc='The series number: [DICOM files] dictionary')


class GroupDicom(BaseInterface):
    input_spec = GroupDicomInputSpec

    output_spec = GroupDicomOutputSpec

    def _run_interface(self, runtime):
        self.grp_dict = group_dicom_files_by_series(*self.inputs.in_files)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['series_files_dict'] = self.grp_dict
        return outputs



import os
import glob
import shutil
from collections import defaultdict
import qixnat
from qiutil.collections import nested_defaultdict
from ... import project


class StagedTestBase(object):

    """
    Base class for testing workflows on a test fixture directory in the
    standard staging subject/session/images hierarchy, e.g.::
        
        Breast003/
            Session01/
                scans/
                    series09.nii.gz
                    series23.nii.gz
                    ...
                breast003_session01_mask.nii.gz
    """

    def __init__(self, logger, fixtures, results, use_mask=False):
        """
        :param logger: this test case's logger
        :param fixtures: the fixtures parent directory
        :param results: the results parent directory
        :param use_mask: flag indicating whether the inputs include
            a mask with the input images
        """
        self._logger = logger
        self._fixtures = fixtures
        self._results = results
        self._use_mask = use_mask

    def setUp(self):
        shutil.rmtree(self._results, True)

    def tearDown(self):
        shutil.rmtree(self._results, True)

    def _test_breast(self, **opts):
        self._test_collection('Breast', **opts)

    def _test_sarcoma(self, **opts):
        self._test_collection('Sarcoma', **opts)

    def _test_collection(self, collection, **opts):
        """
        Run the workflow and verify that the registered images are
        created in XNAT.
        
        :param collection: the collection name
        :param opts: additional workflow options
        """
        # The fixture is the collection subdirectory.
        fixture = os.path.join(self._fixtures, collection.lower())

        # The default workflow base directory.
        if 'base_dir' not in opts:
            opts['base_dir'] = os.path.join(self._results, 'work')

        # The {subject: session: {images: [files], mask: file}} inputs.
        input_dict = self._group_files(fixture)
        # The test subjects.
        subjects = input_dict.keys()
        with qixnat.connect() as xnat:
            # Delete the existing subjects.
            xnat.delete_subjects(project(), *subjects)
            # Iterate over the sessions within subjects.
            for sbj, sess_dict in input_dict.iteritems():
                for sess, sess_opts in sess_dict.iteritems():
                    # The input images.
                    images = sess_opts.pop('images')
                    # The session options now only includes at most the mask.
                    # Add the general workflow options to the session options.
                    sess_opts.update(opts)
                    # Run the workflow in the results work subdirectory.
                    cwd = os.getcwd()
                    work_dir = os.path.join(self._results, 'work')
                    os.makedirs(work_dir)
                    os.chdir(work_dir)
                    try:
                        result = self._run_workflow(sbj, sess, *images, **sess_opts)
                    finally:
                        os.chdir(cwd)
                    # Verify the result.
                    self._verify_result(xnat, sbj, sess, result)
            # Clean up.
            xnat.delete_subjects(project(), *subjects)

    def _group_files(self, fixture):
        """
        Groups the files in the given test fixture directory. The fixture is a
        parent directory which contains a subject/session/scans images hierarchy.
        
        :param fixture: the input data parent directory
        :return: the input *{subject: {session: [images]}}* or
            input *{subject: {session: ([images], mask)}}* dictionary
        """
        # The {subject: {session: [files]}} dictionary return value.
        input_dict = nested_defaultdict(dict, 1)
        # Group the files in each subject subdirectory.
        for sbj_dir in glob.glob(fixture + '/*'):
            _, sbj = os.path.split(sbj_dir)
            self._logger.debug("Discovered test input subject %s in %s" %
                               (sbj, sbj_dir))
            for sess_dir in glob.glob(sbj_dir + '/Session*'):
                _, sess = os.path.split(sess_dir)
                images = glob.glob(sess_dir + '/scans/*')
                if not images:
                    raise ValueError("No images found for %s %s test input in %s"
                                     (sbj, sess, sess_dir))
                self._logger.debug("Discovered %d %s %s test input images in %s" %
                                   (len(images), sbj, sess, sess_dir))
                input_dict[sbj][sess]['images'] = images
                if self._use_mask:
                    masks = glob.glob(sess_dir + '/*mask.*')
                    if not masks:
                        raise ValueError("Mask not found in %s" % sess_dir)
                    if len(masks) > 1:
                        raise ValueError("Too many masks found in %s" % sess_dir)
                    mask = masks[0]
                    self._logger.debug("Discovered %d %s %s test input mask %s" %
                                       (len(images), sbj, sess, mask))
                    input_dict[sbj][sess]['mask'] = mask

        return input_dict

    def _run_workflow(self, subject, session, *images, **opts):
        """
        This method is implemented by subclasses to execute the subclass
        target workflow on the given inputs.
        
        :param subject: the input subject
        :param session: the input session
        :param images: the input image files
        :param opts: the target workflow options
        :return: the execution result
        :raise NotImplementedError: if the class:`StagedTestBase` subclass
            does not implement this method
        """
        raise NotImplementedError("_run_workflow is a subclass responsibility")

    def _verify_result(self, xnat, subject, session, result):
        """
        This method is implemented by subclasses to verify the workflow
        execution result against the given session file inputs. The session
        file inputs parameter is a {(subject, session): *files* dictionary},
        where *files* is a list consisting of the session image files found
        in the test fixture directory.
        
        :param xnat: the :class:`qiutil.qixnats.XNAT` connection
        :param subject: the input subject
        :param session: the input session
        :param result: the :meth`_run_workflow` result
        :raise NotImplementedError: if the class:`StagedTestBase` subclass
            does not implement this method
        """
        raise NotImplementedError(
            "_verify_result is a subclass responsibility")



import os
import re
import glob
import shutil
from nose.tools import (assert_equal, assert_is_not_none)
import nipype.pipeline.engine as pe
from qipipe.pipeline import registration
from ... import (project, ROOT)
from ...helpers.logging import logger
from ...helpers.xnat_test_helper import generate_unique_name
from ...unit.pipeline.staged_test_base import StagedTestBase

REG_CONF = os.path.join(ROOT, 'conf', 'registration.cfg')
"""The test registration configuration."""

FIXTURES = os.path.join(ROOT, 'fixtures', 'registration')
"""The test fixtures directory."""

RESULTS = os.path.join(ROOT, 'results', 'pipeline', 'registration')
"""The test results directory."""

RESOURCE = generate_unique_name(__name__)
"""The XNAT registration resource name."""


class TestRegistrationWorkflow(StagedTestBase):
    """
    Registration workflow unit tests.

    This test exercises the registration workflow on three series of one visit
    in each of the Breast and Sarcoma studies.
    
    The :class:`qiprofile.pipeline.registration.RegistrationWorkflow` technique
    is set to ``mock``, which mocks the registration process by simply copying
    the input scan file to the output "realigned" file. 
    """

    def __init__(self):
        super(TestRegistrationWorkflow, self).__init__(logger(__name__), FIXTURES,
                                                       RESULTS, use_mask=True)

    def test_breast(self):
        super(TestRegistrationWorkflow, self)._test_breast(technique='mock')

    def test_sarcoma(self):
        super(TestRegistrationWorkflow, self)._test_sarcoma(technique='mock')

    def _run_workflow(self, subject, session, *images, **opts):
        """
        Executes :meth:`qipipe.pipeline.registration.run` on the given input.

        :param subject: the input subject
        :param session: the input session
        :param images: the input image files
        :param opts: the :meth:`qipipe.pipeline.registration.run`
            options
        :return: the :meth:`qipipe.pipeline.registration.run` result
        """
        # A reasonable bolus uptake index.
        bolus_arrival_index = min(3, len(images) / 3)
        # The target location.
        dest = os.path.join(RESULTS, subject, session)
        # Execute the workflow.
        return registration.run(subject, session, bolus_arrival_index,
                                cfg_file=REG_CONF, resource=RESOURCE,
                                dest=dest, *images, **opts)

    def _verify_result(self, xnat, subject, session, result):
        """
        :param xnat: the XNAT connection
        :param subject: the registration subject
        :param session: the registration session
        :param result: the meth:`qipipe.pipeline.registration.run` result
            output file paths
        """
        # Verify that the XNAT resource object was created.
        rsc_obj = xnat.find(project(), subject, session, resource=RESOURCE)
        assert_is_not_none(rsc_obj,
                           "The %s %s %s XNAT registration resource object was"
                           " not created" % (subject, session, RESOURCE))

        # Verify that the registration result is accurate.
        dest = os.path.join(RESULTS, subject, session)
        split = (os.path.split(f) for f in result)
        out_dirs, out_files = (set(files) for files in zip(*split))
        rsc_files = set(rsc_obj.files().get())
        assert_equal(out_dirs, set([dest]),
                     "The %s %s %s XNAT registration result directory is"
                      " incorrect - expected %s, found %s" %
                      (subject, session, RESOURCE, dest, out_dirs))
        assert_equal(out_files, rsc_files,
                     "The %s %s %s XNAT registration result file names are"
                     " incorrect - expected %s, found %s" %
                     (subject, session, RESOURCE, rsc_files, out_files))

        # Verify that the output files were created.
        dest_files = (os.path.join(dest, f) for f in os.listdir(dest))
        assert_equal(set(dest_files), set(result),
                     "The %s %s %s XNAT registration result is incorrect:"
                     " %s" % (subject, session, RESOURCE, result))


if __name__ == "__main__":
    import nose

    nose.main(defaultTest=__name__)



## Deprecated - see XNATUpload comment. ##

from nipype.interfaces.base import (
    traits, BaseInterfaceInputSpec, TraitedSpec,
    BaseInterface, InputMultiPath, File)
import qixnat


class XNATUploadInputSpec(BaseInterfaceInputSpec):
    project = traits.Str(mandatory=True, desc='The XNAT project id')

    subject = traits.Str(mandatory=True, desc='The XNAT subject name')

    session = traits.Str(mandatory=True, desc='The XNAT session name')

    resource = traits.Str(desc='The XNAT resource name (scan default is NIFTI)')

    scan = traits.Either(traits.Int, traits.Str, desc='The XNAT scan name')

    reconstruction = traits.Str(desc='The XNAT reconstruction name')

    assessor = traits.Str(desc='The XNAT assessor name')

    force = traits.Bool(desc='Flag indicating whether to replace an existing'
                             ' XNAT file')

    skip_existing = traits.Bool(desc='Flag indicating whether to skip upload'
                                     ' to an existing target XNAT file')

    in_files = InputMultiPath(File(exists=True), mandatory=True,
                              desc='The files to upload')


class XNATUploadOutputSpec(TraitedSpec):
    xnat_files = traits.List(traits.Str, desc='The XNAT file object labels')


class XNATUpload(BaseInterface):
    """
    The ``XNATUpload`` Nipype interface wraps the
    :meth:`qiutil.qixnat.upload` method.
    
    :Note: This XNATUpload interface is deprecated due to the following
        XNAT bug:

        * XNAT or pyxnat concurrent file inserts sporadically fails
          with error that the experiment already exists. Some files
          are inserted, but insert fails unpredictably.

        The work-around is to use the :class:`qipipe.interfaces.XNATCopy`
        interface instead.

    TODO - retry this XNATUpload interface in late 2015 when pyxnat
    hopefully matures.
    """

    input_spec = XNATUploadInputSpec

    output_spec = XNATUploadOutputSpec

    def _run_interface(self, runtime):
        # The upload options.
        opts = {}
        if self.inputs.resource:
            opts['resource'] = self.inputs.resource
        if self.inputs.force:
            opts['force'] = True
        if self.inputs.skip_existing:
            opts['skip_existing'] = True

        # The resource parent.
        if self.inputs.scan:
            opts['modality'] = 'MR'
            opts['scan'] = self.inputs.scan
        elif self.inputs.reconstruction:
            opts['reconstruction'] = self.inputs.reconstruction
        elif self.inputs.assessor:
            opts['assessor'] = self.inputs.assessor

        # Upload the files.
        with qixnat.connect() as xnat:
            self._xnat_files = xnat.upload(
                self.inputs.project, self.inputs.subject, self.inputs.session,
                *self.inputs.in_files, **opts)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        if hasattr(self, '_xnat_files'):
            outputs['xnat_files'] = self._xnat_files

        return outputs



import os
import shutil
from nose.tools import assert_true
from qipipe.pipeline import staging
import qixnat
from qipipe.staging.staging_helper import (get_subjects, iter_stage)
from ... import (project, ROOT)
from ...helpers.logging import logger


# FIXME - test module resolves to qiutil/test!?!
# This, and presumably other, tests are consequently broken.
# Relative import fails.
# TODO - WHY DOES THIS HAPPEN? HOW TO FIX IT?


FIXTURES = os.path.join(ROOT, 'fixtures', 'staging')
"""The test fixture directory."""

RESULTS = os.path.join(ROOT, 'results', 'pipeline', 'staging')
"""The test results directory."""


class TestStagingWorkflow(object):
    """Staging workflow unit tests."""

    def setUp(self):
        shutil.rmtree(RESULTS, True)

    def tearDown(self):
        shutil.rmtree(RESULTS, True)

    def test_breast(self):
        self._test_collection('Breast')

    def test_t2(self):
        self._test_collection('Breast')

    def test_sarcoma(self):
        self._test_collection('Sarcoma')

    def _test_collection(self, collection):
        """
        Run the staging workflow on the given collection and verify that
        the sessions are created in XNAT.

        Note:: This test does not verify the CTP staging area nor that the
        image files are correctly uploaded. These features should be
        verified manually.
        
        :param collection: the AIRC collection name
        """
        fixture = os.path.join(FIXTURES, collection.lower())
        logger(__name__).debug(
            "Testing the staging workflow on %s..." % fixture)

        # The staging destination and work area.
        dest = os.path.join(RESULTS, 'staged')
        work = os.path.join(RESULTS, 'work')

        # The test subject => directory dictionary.
        sbj_dir_dict = get_subjects(collection, fixture)
        # The test subjects.
        subjects = sbj_dir_dict.keys()
        # The test source directories.
        inputs = sbj_dir_dict.values()

        with qixnat.connect() as xnat:
            # Delete any existing test subjects.
            xnat.delete_subjects(project(), *subjects)
            # Run the workflow on each session fixture.
            for sbj, sess, scan_type_dict in iter_stage(collection, *inputs,
                                                        dest=dest):
                for scan_type, scan_dict in scan_type_dict.iteritems():
                    work_dir = os.path.join(work, scan_type)
                    stg_wf = staging.StagingWorkflow(scan_type,
                                                     base_dir=work_dir)
                    stg_wf.set_inputs(collection, sbj, sess, scan_dict,
                                      dest=dest)
                    stg_wf.run()
                    # Verify the result.
                    sess_obj = xnat.get_session(project(), sbj, sess)
                    assert_true(sess_obj.exists(),
                                "The %s %s session was not created in XNAT" %
                                (sbj, sess))
                    sess_dest = os.path.join(dest, sbj, sess)
                    assert_true(os.path.exists(sess_dest), "The staging area"
                                " was not created: %s" % sess_dest)
                    for scan, dcm_files in scan_dict.iteritems():
                        scan_obj = xnat.get_scan(project(), sbj, sess, scan)
                        assert_true(scan_obj.exists(), "The %s %s scan %s was"
                                    " not created in XNAT" % (sbj, sess, scan))

            # Delete the test subjects.
            xnat.delete_subjects(project(), *subjects)


if __name__ == "__main__":
    import nose

    nose.main(defaultTest=__name__)



def get_available_models():
    """
    @return a mock result that permits building, but not running, a workflow
    """
    class MockModel(object):
        def __init__(self):
            self.optimization_params = ['tau_i', 'chisq', 'k_trans', 'v_e', 'chisq']
    
    mock_mdl = MockModel()
    
    return [{'fxr.model': mock_mdl}, None, None]
